"""Provide event models for the Home Connect API."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import StrEnum
import json

from httpx_sse import ServerSentEvent
from mashumaro import field_options
from mashumaro.mixins.json import DataClassJSONMixin

from aiohomeconnect.const import LOGGER
from aiohomeconnect.model.program import OptionKey
from aiohomeconnect.model.setting import SettingKey
from aiohomeconnect.model.status import StatusKey


@dataclass
class ArrayOfEvents(DataClassJSONMixin):
    """Represent ArrayOfEvents."""

    items: list[Event]


@dataclass
class Event(DataClassJSONMixin):
    """Represent Event."""

    key: EventKey
    raw_key: str = field(metadata=field_options(alias="key"))
    timestamp: int
    level: str
    handling: str
    value: str | int | float | bool | None
    name: str | None = None
    uri: str | None = None
    display_value: str | None = field(
        default=None, metadata=field_options(alias="displayvalue")
    )
    unit: str | None = None


@dataclass
class EventMessage:
    """Represent a server sent event message sent from the Home Connect API."""

    ha_id: str
    type: EventType
    data: ArrayOfEvents

    @classmethod
    def from_server_sent_event(cls, sse: ServerSentEvent) -> EventMessage:
        """Create an EventMessage instance from a server sent event."""
        if not sse.data:
            return cls(
                ha_id=sse.id,
                type=EventType(sse.event),
                data=ArrayOfEvents([]),
            )
        data = json.loads(sse.data)
        if "items" in data:
            events = ArrayOfEvents.from_dict(data)
        else:
            events = ArrayOfEvents([Event.from_dict(data)])
        return cls(
            ha_id=sse.id,
            type=EventType(sse.event),
            data=events,
        )


class EventKey(StrEnum):
    """Represent an event key."""

    @classmethod
    def _missing_(cls, value: object) -> EventKey:
        """Return UNKNOWN for missing keys."""
        LOGGER.debug("Unknown event key: %s", value)
        return cls.UNKNOWN

    UNKNOWN = "unknown"
    BSH_COMMON_APPLIANCE_CONNECTED = "BSH.Common.Appliance.Connected"
    BSH_COMMON_APPLIANCE_DEPAIRED = "BSH.Common.Appliance.Depaired"
    BSH_COMMON_APPLIANCE_DISCONNECTED = "BSH.Common.Appliance.Disconnected"
    BSH_COMMON_APPLIANCE_PAIRED = "BSH.Common.Appliance.Paired"
    BSH_COMMON_EVENT_ALARM_CLOCK_ELAPSED = "BSH.Common.Event.AlarmClockElapsed"
    BSH_COMMON_EVENT_PROGRAM_ABORTED = "BSH.Common.Event.ProgramAborted"
    BSH_COMMON_EVENT_PROGRAM_FINISHED = "BSH.Common.Event.ProgramFinished"
    BSH_COMMON_OPTION_DURATION = OptionKey.BSH_COMMON_DURATION
    BSH_COMMON_OPTION_ELAPSED_PROGRAM_TIME = "BSH.Common.Option.ElapsedProgramTime"
    BSH_COMMON_OPTION_ESTIMATED_TOTAL_PROGRAM_TIME = (
        OptionKey.BSH_COMMON_ESTIMATED_TOTAL_PROGRAM_TIME
    )
    BSH_COMMON_OPTION_FINISH_IN_RELATIVE = OptionKey.BSH_COMMON_FINISH_IN_RELATIVE
    BSH_COMMON_OPTION_PROGRAM_PROGRESS = "BSH.Common.Option.ProgramProgress"
    BSH_COMMON_OPTION_REMAINING_PROGRAM_TIME = "BSH.Common.Option.RemainingProgramTime"
    BSH_COMMON_OPTION_REMAINING_PROGRAM_TIME_IS_ESTIMATED = (
        OptionKey.BSH_COMMON_REMAINING_PROGRAM_TIME_IS_ESTIMATED
    )
    BSH_COMMON_OPTION_START_IN_RELATIVE = OptionKey.BSH_COMMON_START_IN_RELATIVE
    BSH_COMMON_ROOT_ACTIVE_PROGRAM = "BSH.Common.Root.ActiveProgram"
    BSH_COMMON_ROOT_SELECTED_PROGRAM = "BSH.Common.Root.SelectedProgram"
    BSH_COMMON_SETTING_ALARM_CLOCK = SettingKey.BSH_COMMON_ALARM_CLOCK
    BSH_COMMON_SETTING_AMBIENT_LIGHT_BRIGHTNESS = (
        SettingKey.BSH_COMMON_AMBIENT_LIGHT_BRIGHTNESS
    )
    BSH_COMMON_SETTING_AMBIENT_LIGHT_COLOR = SettingKey.BSH_COMMON_AMBIENT_LIGHT_COLOR
    BSH_COMMON_SETTING_AMBIENT_LIGHT_CUSTOM_COLOR = (
        SettingKey.BSH_COMMON_AMBIENT_LIGHT_CUSTOM_COLOR
    )
    BSH_COMMON_SETTING_AMBIENT_LIGHT_ENABLED = (
        SettingKey.BSH_COMMON_AMBIENT_LIGHT_ENABLED
    )
    BSH_COMMON_SETTING_CHILD_LOCK = SettingKey.BSH_COMMON_CHILD_LOCK
    BSH_COMMON_SETTING_LIQUID_VOLUME_UNIT = SettingKey.BSH_COMMON_LIQUID_VOLUME_UNIT
    BSH_COMMON_SETTING_POWER_STATE = SettingKey.BSH_COMMON_POWER_STATE
    BSH_COMMON_SETTING_TEMPERATURE_UNIT = SettingKey.BSH_COMMON_TEMPERATURE_UNIT
    BSH_COMMON_STATUS_BATTERY_CHARGING_STATE = (
        StatusKey.BSH_COMMON_BATTERY_CHARGING_STATE
    )
    BSH_COMMON_STATUS_BATTERY_LEVEL = StatusKey.BSH_COMMON_BATTERY_LEVEL
    BSH_COMMON_STATUS_CHARGING_CONNECTION = StatusKey.BSH_COMMON_CHARGING_CONNECTION
    BSH_COMMON_STATUS_DOOR_STATE = StatusKey.BSH_COMMON_DOOR_STATE
    BSH_COMMON_STATUS_INTERIOR_ILLUMINATION_ACTIVE = (
        StatusKey.BSH_COMMON_INTERIOR_ILLUMINATION_ACTIVE
    )
    BSH_COMMON_STATUS_LOCAL_CONTROL_ACTIVE = StatusKey.BSH_COMMON_LOCAL_CONTROL_ACTIVE
    BSH_COMMON_STATUS_OPERATION_STATE = StatusKey.BSH_COMMON_OPERATION_STATE
    BSH_COMMON_STATUS_REMOTE_CONTROL_ACTIVE = StatusKey.BSH_COMMON_REMOTE_CONTROL_ACTIVE
    BSH_COMMON_STATUS_REMOTE_CONTROL_START_ALLOWED = (
        StatusKey.BSH_COMMON_REMOTE_CONTROL_START_ALLOWED
    )
    BSH_COMMON_STATUS_VIDEO_CAMERA_STATE = StatusKey.BSH_COMMON_VIDEO_CAMERA_STATE
    CONSUMER_PRODUCTS_CLEANING_ROBOT_EVENT_DOCKING_STATION_NOT_FOUND = (
        "ConsumerProducts.CleaningRobot.Event.DockingStationNotFound"
    )
    CONSUMER_PRODUCTS_CLEANING_ROBOT_EVENT_EMPTY_DUST_BOX_AND_CLEAN_FILTER = (
        "ConsumerProducts.CleaningRobot.Event.EmptyDustBoxAndCleanFilter"
    )
    CONSUMER_PRODUCTS_CLEANING_ROBOT_EVENT_ROBOT_IS_STUCK = (
        "ConsumerProducts.CleaningRobot.Event.RobotIsStuck"
    )
    CONSUMER_PRODUCTS_CLEANING_ROBOT_OPTION_CLEANING_MODE = (
        OptionKey.CONSUMER_PRODUCTS_CLEANING_ROBOT_CLEANING_MODE
    )
    CONSUMER_PRODUCTS_CLEANING_ROBOT_OPTION_PROCESS_PHASE = (
        "ConsumerProducts.CleaningRobot.Option.ProcessPhase"
    )
    CONSUMER_PRODUCTS_CLEANING_ROBOT_OPTION_REFERENCE_MAP_ID = (
        OptionKey.CONSUMER_PRODUCTS_CLEANING_ROBOT_REFERENCE_MAP_ID
    )
    CONSUMER_PRODUCTS_CLEANING_ROBOT_OPTION_SUCTION_POWER = (
        OptionKey.CONSUMER_PRODUCTS_CLEANING_ROBOT_SUCTION_POWER
    )
    CONSUMER_PRODUCTS_CLEANING_ROBOT_SETTING_CURRENT_MAP = (
        SettingKey.CONSUMER_PRODUCTS_CLEANING_ROBOT_CURRENT_MAP
    )
    CONSUMER_PRODUCTS_CLEANING_ROBOT_SETTING_NAME_OF_MAP_1 = (
        SettingKey.CONSUMER_PRODUCTS_CLEANING_ROBOT_NAME_OF_MAP_1
    )
    CONSUMER_PRODUCTS_CLEANING_ROBOT_SETTING_NAME_OF_MAP_2 = (
        SettingKey.CONSUMER_PRODUCTS_CLEANING_ROBOT_NAME_OF_MAP_2
    )
    CONSUMER_PRODUCTS_CLEANING_ROBOT_SETTING_NAME_OF_MAP_3 = (
        SettingKey.CONSUMER_PRODUCTS_CLEANING_ROBOT_NAME_OF_MAP_3
    )
    CONSUMER_PRODUCTS_CLEANING_ROBOT_SETTING_NAME_OF_MAP_4 = (
        SettingKey.CONSUMER_PRODUCTS_CLEANING_ROBOT_NAME_OF_MAP_4
    )
    CONSUMER_PRODUCTS_CLEANING_ROBOT_SETTING_NAME_OF_MAP_5 = (
        SettingKey.CONSUMER_PRODUCTS_CLEANING_ROBOT_NAME_OF_MAP_5
    )
    CONSUMER_PRODUCTS_CLEANING_ROBOT_STATUS_DUST_BOX_INSERTED = (
        StatusKey.CONSUMER_PRODUCTS_CLEANING_ROBOT_DUST_BOX_INSERTED
    )
    CONSUMER_PRODUCTS_CLEANING_ROBOT_STATUS_LAST_SELECTED_MAP = (
        StatusKey.CONSUMER_PRODUCTS_CLEANING_ROBOT_LAST_SELECTED_MAP
    )
    CONSUMER_PRODUCTS_CLEANING_ROBOT_STATUS_LIFTED = (
        StatusKey.CONSUMER_PRODUCTS_CLEANING_ROBOT_LIFTED
    )
    CONSUMER_PRODUCTS_CLEANING_ROBOT_STATUS_LOST = (
        StatusKey.CONSUMER_PRODUCTS_CLEANING_ROBOT_LOST
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_BEAN_CONTAINER_EMPTY = (
        "ConsumerProducts.CoffeeMaker.Event.BeanContainerEmpty"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_CALC_N_CLEAN_IN10CUPS = (
        "ConsumerProducts.CoffeeMaker.Event.CalcNCleanIn10Cups"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_CALC_N_CLEAN_IN15CUPS = (
        "ConsumerProducts.CoffeeMaker.Event.CalcNCleanIn15Cups"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_CALC_N_CLEAN_IN20CUPS = (
        "ConsumerProducts.CoffeeMaker.Event.CalcNCleanIn20Cups"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_CALC_N_CLEAN_IN5CUPS = (
        "ConsumerProducts.CoffeeMaker.Event.CalcNCleanIn5Cup"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_DESCALING_IN_10_CUPS = (
        "ConsumerProducts.CoffeeMaker.Event.DescalingIn10Cups"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_DESCALING_IN_15_CUPS = (
        "ConsumerProducts.CoffeeMaker.Event.DescalingIn15Cups"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_DESCALING_IN_20_CUPS = (
        "ConsumerProducts.CoffeeMaker.Event.DescalingIn20Cups"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_DESCALING_IN_5_CUPS = (
        "ConsumerProducts.CoffeeMaker.Event.DescalingIn5Cups"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_DEVICE_CALC_N_CLEAN_BLOCKAGE = (
        "ConsumerProducts.CoffeeMaker.Event.DeviceCalcNCleanBlockage"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_DEVICE_CALC_N_CLEAN_OVERDUE = (
        "ConsumerProducts.CoffeeMaker.Event.DeviceCalcNCleanOverdue"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_DEVICE_CLEANING_OVERDUE = (
        "ConsumerProducts.CoffeeMaker.Event.DeviceCleaningOverdue"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_DEVICE_DESCALING_BLOCKAGE = (
        "ConsumerProducts.CoffeeMaker.Event.DeviceDescalingBlockage"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_DEVICE_DESCALING_OVERDUE = (
        "ConsumerProducts.CoffeeMaker.Event.DeviceDescalingOverdue"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_DEVICE_SHOULD_BE_CALC_N_CLEANED = (
        "ConsumerProducts.CoffeeMaker.Event.DeviceShouldBeCalcNCleaned"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_DEVICE_SHOULD_BE_CLEANED = (
        "ConsumerProducts.CoffeeMaker.Event.DeviceShouldBeCleaned"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_DEVICE_SHOULD_BE_DESCALED = (
        "ConsumerProducts.CoffeeMaker.Event.DeviceShouldBeDescaled"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_DRIP_TRAY_FULL = (
        "ConsumerProducts.CoffeeMaker.Event.DripTrayFull"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_KEEP_MILK_TANK_COOL = (
        "ConsumerProducts.CoffeeMaker.Event.KeepMilkTankCool"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_EVENT_WATER_TANK_EMPTY = (
        "ConsumerProducts.CoffeeMaker.Event.WaterTankEmpty"
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_OPTION_BEAN_AMOUNT = (
        OptionKey.CONSUMER_PRODUCTS_COFFEE_MAKER_BEAN_AMOUNT
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_OPTION_BEAN_CONTAINER_SELECTION = (
        OptionKey.CONSUMER_PRODUCTS_COFFEE_MAKER_BEAN_CONTAINER_SELECTION
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_OPTION_COFFEE_MILK_RATIO = (
        OptionKey.CONSUMER_PRODUCTS_COFFEE_MAKER_COFFEE_MILK_RATIO
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_OPTION_COFFEE_TEMPERATURE = (
        OptionKey.CONSUMER_PRODUCTS_COFFEE_MAKER_COFFEE_TEMPERATURE
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_OPTION_FILL_QUANTITY = (
        OptionKey.CONSUMER_PRODUCTS_COFFEE_MAKER_FILL_QUANTITY
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_OPTION_FLOW_RATE = (
        OptionKey.CONSUMER_PRODUCTS_COFFEE_MAKER_FLOW_RATE
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_OPTION_HOT_WATER_TEMPERATURE = (
        OptionKey.CONSUMER_PRODUCTS_COFFEE_MAKER_HOT_WATER_TEMPERATURE
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_OPTION_MULTIPLE_BEVERAGES = (
        OptionKey.CONSUMER_PRODUCTS_COFFEE_MAKER_MULTIPLE_BEVERAGES
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_SETTING_CUP_WARMER = (
        SettingKey.CONSUMER_PRODUCTS_COFFEE_MAKER_CUP_WARMER
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_STATUS_BEVERAGE_COUNTER_COFFEE = (
        StatusKey.CONSUMER_PRODUCTS_COFFEE_MAKER_BEVERAGE_COUNTER_COFFEE
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_STATUS_BEVERAGE_COUNTER_COFFEE_AND_MILK = (
        StatusKey.CONSUMER_PRODUCTS_COFFEE_MAKER_BEVERAGE_COUNTER_COFFEE_AND_MILK
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_STATUS_BEVERAGE_COUNTER_FROTHY_MILK = (
        StatusKey.CONSUMER_PRODUCTS_COFFEE_MAKER_BEVERAGE_COUNTER_FROTHY_MILK
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_STATUS_BEVERAGE_COUNTER_HOT_MILK = (
        StatusKey.CONSUMER_PRODUCTS_COFFEE_MAKER_BEVERAGE_COUNTER_HOT_MILK
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_STATUS_BEVERAGE_COUNTER_HOT_WATER = (
        StatusKey.CONSUMER_PRODUCTS_COFFEE_MAKER_BEVERAGE_COUNTER_HOT_WATER
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_STATUS_BEVERAGE_COUNTER_HOT_WATER_CUPS = (
        StatusKey.CONSUMER_PRODUCTS_COFFEE_MAKER_BEVERAGE_COUNTER_HOT_WATER_CUPS
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_STATUS_BEVERAGE_COUNTER_MILK = (
        StatusKey.CONSUMER_PRODUCTS_COFFEE_MAKER_BEVERAGE_COUNTER_MILK
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_STATUS_BEVERAGE_COUNTER_POWDER_COFFEE = (
        StatusKey.CONSUMER_PRODUCTS_COFFEE_MAKER_BEVERAGE_COUNTER_POWDER_COFFEE
    )
    CONSUMER_PRODUCTS_COFFEE_MAKER_STATUS_BEVERAGE_COUNTER_RISTRETTO_ESPRESSO = (
        StatusKey.CONSUMER_PRODUCTS_COFFEE_MAKER_BEVERAGE_COUNTER_RISTRETTO_ESPRESSO
    )
    COOKING_COMMON_EVENT_HOOD_GREASE_FILTER_MAX_SATURATION_NEARLY_REACHED = (
        "Cooking.Common.Event.Hood.GreaseFilterMaxSaturationNearlyReached"
    )
    COOKING_COMMON_EVENT_HOOD_GREASE_FILTER_MAX_SATURATION_REACHED = (
        "Cooking.Common.Event.Hood.GreaseFilterMaxSaturationReached"
    )
    COOKING_COMMON_OPTION_HOOD_INTENSIVE_LEVEL = (
        OptionKey.COOKING_COMMON_HOOD_INTENSIVE_LEVEL
    )
    COOKING_COMMON_OPTION_HOOD_VENTING_LEVEL = (
        OptionKey.COOKING_COMMON_HOOD_VENTING_LEVEL
    )
    COOKING_COMMON_SETTING_LIGHTING = SettingKey.COOKING_COMMON_LIGHTING
    COOKING_COMMON_SETTING_LIGHTING_BRIGHTNESS = (
        SettingKey.COOKING_COMMON_LIGHTING_BRIGHTNESS
    )
    COOKING_HOOD_SETTING_COLOR_TEMPERATURE = SettingKey.COOKING_HOOD_COLOR_TEMPERATURE
    COOKING_HOOD_SETTING_COLOR_TEMPERATURE_PERCENT = (
        SettingKey.COOKING_HOOD_COLOR_TEMPERATURE_PERCENT
    )
    COOKING_OVEN_EVENT_PREHEAT_FINISHED = "Cooking.Oven.Event.PreheatFinished"
    COOKING_OVEN_EVENT_REGULAR_PREHEAT_FINISHED = (
        "Cooking.Oven.Event.RegularPreheatFinished"
    )
    COOKING_OVEN_OPTION_FAST_PRE_HEAT = OptionKey.COOKING_OVEN_FAST_PRE_HEAT
    COOKING_OVEN_OPTION_SETPOINT_TEMPERATURE = (
        OptionKey.COOKING_OVEN_SETPOINT_TEMPERATURE
    )
    COOKING_OVEN_OPTION_WARMING_LEVEL = OptionKey.COOKING_OVEN_WARMING_LEVEL
    COOKING_OVEN_SETTING_SABBATH_MODE = SettingKey.COOKING_OVEN_SABBATH_MODE
    COOKING_OVEN_STATUS_CURRENT_CAVITY_TEMPERATURE = (
        StatusKey.COOKING_OVEN_CURRENT_CAVITY_TEMPERATURE
    )
    DISHCARE_DISHWASHER_EVENT_RINSE_AID_NEARLY_EMPTY = (
        "Dishcare.Dishwasher.Event.RinseAidNearlyEmpty"
    )
    DISHCARE_DISHWASHER_EVENT_SALT_NEARLY_EMPTY = (
        "Dishcare.Dishwasher.Event.SaltNearlyEmpty"
    )
    DISHCARE_DISHWASHER_OPTION_BRILLIANCE_DRY = (
        OptionKey.DISHCARE_DISHWASHER_BRILLIANCE_DRY
    )
    DISHCARE_DISHWASHER_OPTION_ECO_DRY = OptionKey.DISHCARE_DISHWASHER_ECO_DRY
    DISHCARE_DISHWASHER_OPTION_EXTRA_DRY = OptionKey.DISHCARE_DISHWASHER_EXTRA_DRY
    DISHCARE_DISHWASHER_OPTION_HALF_LOAD = OptionKey.DISHCARE_DISHWASHER_HALF_LOAD
    DISHCARE_DISHWASHER_OPTION_HYGIENE_PLUS = OptionKey.DISHCARE_DISHWASHER_HYGIENE_PLUS
    DISHCARE_DISHWASHER_OPTION_INTENSIV_ZONE = (
        OptionKey.DISHCARE_DISHWASHER_INTENSIV_ZONE
    )
    DISHCARE_DISHWASHER_OPTION_SILENCE_ON_DEMAND = (
        OptionKey.DISHCARE_DISHWASHER_SILENCE_ON_DEMAND
    )
    DISHCARE_DISHWASHER_OPTION_VARIO_SPEED_PLUS = (
        OptionKey.DISHCARE_DISHWASHER_VARIO_SPEED_PLUS
    )
    DISHCARE_DISHWASHER_OPTION_ZEOLITE_DRY = OptionKey.DISHCARE_DISHWASHER_ZEOLITE_DRY
    HEATING_VENTILATION_AIR_CONDITIONING_AIR_CONDITIONER_FAN_SPEED_MODE = (
        OptionKey.HEATING_VENTILATION_AIR_CONDITIONING_AIR_CONDITIONER_FAN_SPEED_MODE
    )
    HEATING_VENTILATION_AIR_CONDITIONING_AIR_CONDITIONER_FAN_SPEED_PERCENTAGE = (
        OptionKey.HEATING_VENTILATION_AIR_CONDITIONING_AIR_CONDITIONER_FAN_SPEED_PERCENTAGE  # noqa: E501, RUF100
    )
    LAUNDRY_CARE_COMMON_OPTION_LOAD_RECOMMENDATION = (
        OptionKey.LAUNDRY_CARE_COMMON_LOAD_RECOMMENDATION
    )
    LAUNDRY_CARE_COMMON_OPTION_VARIO_PERFECT = (
        OptionKey.LAUNDRY_CARE_COMMON_VARIO_PERFECT
    )
    LAUNDRY_CARE_COMMON_OPTION_SILENT_MODE = OptionKey.LAUNDRY_CARE_COMMON_SILENT_MODE
    LAUNDRY_CARE_DRYER_EVENT_DRYING_PROCESS_FINISHED = (
        "LaundryCare.Dryer.Event.DryingProcessFinished"
    )
    LAUNDRY_CARE_DRYER_OPTION_DRYING_TARGET = OptionKey.LAUNDRY_CARE_DRYER_DRYING_TARGET
    LAUNDRY_CARE_WASHER_EVENT_I_DOS_1_FILL_LEVEL_POOR = (
        "LaundryCare.Washer.Event.IDos1FillLevelPoor"
    )
    LAUNDRY_CARE_WASHER_EVENT_I_DOS_2_FILL_LEVEL_POOR = (
        "LaundryCare.Washer.Event.IDos2FillLevelPoor"
    )
    LAUNDRY_CARE_WASHER_OPTION_I_DOS_1_ACTIVE = (
        OptionKey.LAUNDRY_CARE_WASHER_I_DOS_1_ACTIVE
    )
    LAUNDRY_CARE_WASHER_OPTION_I_DOS_2_ACTIVE = (
        OptionKey.LAUNDRY_CARE_WASHER_I_DOS_2_ACTIVE
    )
    LAUNDRY_CARE_WASHER_OPTION_INTENSIVE_PLUS = (
        OptionKey.LAUNDRY_CARE_WASHER_INTENSIVE_PLUS
    )
    LAUNDRY_CARE_WASHER_OPTION_LESS_IRONING = OptionKey.LAUNDRY_CARE_WASHER_LESS_IRONING
    LAUNDRY_CARE_WASHER_OPTION_MINI_LOAD = OptionKey.LAUNDRY_CARE_WASHER_MINI_LOAD
    LAUNDRY_CARE_WASHER_OPTION_PREWASH = OptionKey.LAUNDRY_CARE_WASHER_PREWASH
    LAUNDRY_CARE_WASHER_OPTION_RINSE_HOLD = OptionKey.LAUNDRY_CARE_WASHER_RINSE_HOLD
    LAUNDRY_CARE_WASHER_OPTION_RINSE_PLUS = OptionKey.LAUNDRY_CARE_WASHER_RINSE_PLUS
    LAUNDRY_CARE_WASHER_OPTION_SOAK = OptionKey.LAUNDRY_CARE_WASHER_SOAK
    LAUNDRY_CARE_WASHER_OPTION_SPIN_SPEED = OptionKey.LAUNDRY_CARE_WASHER_SPIN_SPEED
    LAUNDRY_CARE_WASHER_OPTION_STAINS = OptionKey.LAUNDRY_CARE_WASHER_STAINS
    LAUNDRY_CARE_WASHER_OPTION_WATER_PLUS = OptionKey.LAUNDRY_CARE_WASHER_WATER_PLUS
    LAUNDRY_CARE_WASHER_OPTION_TEMPERATURE = OptionKey.LAUNDRY_CARE_WASHER_TEMPERATURE
    LAUNDRY_CARE_WASHER_SETTING_I_DOS_1_BASE_LEVEL = (
        SettingKey.LAUNDRY_CARE_WASHER_I_DOS_1_BASE_LEVEL
    )
    LAUNDRY_CARE_WASHER_SETTING_I_DOS_2_BASE_LEVEL = (
        SettingKey.LAUNDRY_CARE_WASHER_I_DOS_2_BASE_LEVEL
    )
    REFRIGERATION_COMMON_SETTING_BOTTLE_COOLER_SETPOINT_TEMPERATURE = (
        SettingKey.REFRIGERATION_COMMON_BOTTLE_COOLER_SETPOINT_TEMPERATURE
    )
    REFRIGERATION_COMMON_SETTING_CHILLER_COMMON_SETPOINT_TEMPERATURE = (
        SettingKey.REFRIGERATION_COMMON_CHILLER_COMMON_SETPOINT_TEMPERATURE
    )
    REFRIGERATION_COMMON_SETTING_CHILLER_LEFT_SETPOINT_TEMPERATURE = (
        SettingKey.REFRIGERATION_COMMON_CHILLER_LEFT_SETPOINT_TEMPERATURE
    )
    REFRIGERATION_COMMON_SETTING_CHILLER_RIGHT_SETPOINT_TEMPERATURE = (
        SettingKey.REFRIGERATION_COMMON_CHILLER_RIGHT_SETPOINT_TEMPERATURE
    )
    REFRIGERATION_COMMON_SETTING_DISPENSER_ENABLED = (
        SettingKey.REFRIGERATION_COMMON_DISPENSER_ENABLED
    )
    REFRIGERATION_COMMON_SETTING_DOOR_ASSISTANT_FORCE_FREEZER = (
        SettingKey.REFRIGERATION_COMMON_DOOR_ASSISTANT_FORCE_FREEZER
    )
    REFRIGERATION_COMMON_SETTING_DOOR_ASSISTANT_FORCE_FRIDGE = (
        SettingKey.REFRIGERATION_COMMON_DOOR_ASSISTANT_FORCE_FRIDGE
    )
    REFRIGERATION_COMMON_SETTING_DOOR_ASSISTANT_FREEZER = (
        SettingKey.REFRIGERATION_COMMON_DOOR_ASSISTANT_FREEZER
    )
    REFRIGERATION_COMMON_SETTING_DOOR_ASSISTANT_FRIDGE = (
        SettingKey.REFRIGERATION_COMMON_DOOR_ASSISTANT_FRIDGE
    )
    REFRIGERATION_COMMON_SETTING_DOOR_ASSISTANT_TIMEOUT_FREEZER = (
        SettingKey.REFRIGERATION_COMMON_DOOR_ASSISTANT_TIMEOUT_FREEZER
    )
    REFRIGERATION_COMMON_SETTING_DOOR_ASSISTANT_TIMEOUT_FRIDGE = (
        SettingKey.REFRIGERATION_COMMON_DOOR_ASSISTANT_TIMEOUT_FRIDGE
    )
    REFRIGERATION_COMMON_SETTING_DOOR_ASSISTANT_TRIGGER_FREEZER = (
        SettingKey.REFRIGERATION_COMMON_DOOR_ASSISTANT_TRIGGER_FREEZER
    )
    REFRIGERATION_COMMON_SETTING_DOOR_ASSISTANT_TRIGGER_FRIDGE = (
        SettingKey.REFRIGERATION_COMMON_DOOR_ASSISTANT_TRIGGER_FRIDGE
    )
    REFRIGERATION_COMMON_SETTING_ECO_MODE = SettingKey.REFRIGERATION_COMMON_ECO_MODE
    REFRIGERATION_COMMON_SETTING_FRESH_MODE = SettingKey.REFRIGERATION_COMMON_FRESH_MODE
    REFRIGERATION_COMMON_SETTING_LIGHT_EXTERNAL_BRIGHTNESS = (
        SettingKey.REFRIGERATION_COMMON_LIGHT_EXTERNAL_BRIGHTNESS
    )
    REFRIGERATION_COMMON_SETTING_LIGHT_EXTERNAL_POWER = (
        SettingKey.REFRIGERATION_COMMON_LIGHT_EXTERNAL_POWER
    )
    REFRIGERATION_COMMON_SETTING_LIGHT_INTERNAL_BRIGHTNESS = (
        SettingKey.REFRIGERATION_COMMON_LIGHT_INTERNAL_BRIGHTNESS
    )
    REFRIGERATION_COMMON_SETTING_LIGHT_INTERNAL_POWER = (
        SettingKey.REFRIGERATION_COMMON_LIGHT_INTERNAL_POWER
    )
    REFRIGERATION_COMMON_SETTING_SABBATH_MODE = (
        SettingKey.REFRIGERATION_COMMON_SABBATH_MODE
    )
    REFRIGERATION_COMMON_SETTING_VACATION_MODE = (
        SettingKey.REFRIGERATION_COMMON_VACATION_MODE
    )
    REFRIGERATION_COMMON_SETTING_WINE_COMPARTMENT_2_SETPOINT_TEMPERATURE = (
        SettingKey.REFRIGERATION_COMMON_WINE_COMPARTMENT_2_SETPOINT_TEMPERATURE
    )
    REFRIGERATION_COMMON_SETTING_WINE_COMPARTMENT_3_SETPOINT_TEMPERATURE = (
        SettingKey.REFRIGERATION_COMMON_WINE_COMPARTMENT_3_SETPOINT_TEMPERATURE
    )
    REFRIGERATION_COMMON_SETTING_WINE_COMPARTMENT_SETPOINT_TEMPERATURE = (
        SettingKey.REFRIGERATION_COMMON_WINE_COMPARTMENT_SETPOINT_TEMPERATURE
    )
    REFRIGERATION_COMMON_STATUS_DOOR_BOTTLE_COOLER = (
        StatusKey.REFRIGERATION_COMMON_DOOR_BOTTLE_COOLER
    )
    REFRIGERATION_COMMON_STATUS_DOOR_CHILLER = (
        StatusKey.REFRIGERATION_COMMON_DOOR_CHILLER
    )
    REFRIGERATION_COMMON_STATUS_DOOR_CHILLER_COMMON = (
        StatusKey.REFRIGERATION_COMMON_DOOR_CHILLER_COMMON
    )
    REFRIGERATION_COMMON_STATUS_DOOR_CHILLER_LEFT = (
        StatusKey.REFRIGERATION_COMMON_DOOR_CHILLER_LEFT
    )
    REFRIGERATION_COMMON_STATUS_DOOR_CHILLER_RIGHT = (
        StatusKey.REFRIGERATION_COMMON_DOOR_CHILLER_RIGHT
    )
    REFRIGERATION_COMMON_STATUS_DOOR_FLEX_COMPARTMENT = (
        StatusKey.REFRIGERATION_COMMON_DOOR_FLEX_COMPARTMENT
    )
    REFRIGERATION_COMMON_STATUS_DOOR_FREEZER = (
        StatusKey.REFRIGERATION_COMMON_DOOR_FREEZER
    )
    REFRIGERATION_COMMON_STATUS_DOOR_REFRIGERATOR = (
        StatusKey.REFRIGERATION_COMMON_DOOR_REFRIGERATOR
    )
    REFRIGERATION_COMMON_STATUS_DOOR_REFRIGERATOR2 = (
        StatusKey.REFRIGERATION_COMMON_DOOR_REFRIGERATOR_2
    )
    REFRIGERATION_COMMON_STATUS_DOOR_REFRIGERATOR3 = (
        StatusKey.REFRIGERATION_COMMON_DOOR_REFRIGERATOR_3
    )
    REFRIGERATION_COMMON_STATUS_DOOR_WINE_COMPARTMENT = (
        StatusKey.REFRIGERATION_COMMON_DOOR_WINE_COMPARTMENT
    )
    REFRIGERATION_FRIDGE_FREEZER_EVENT_DOOR_ALARM_FREEZER = (
        "Refrigeration.FridgeFreezer.Event.DoorAlarmFreezer"
    )
    REFRIGERATION_FRIDGE_FREEZER_EVENT_DOOR_ALARM_REFRIGERATOR = (
        "Refrigeration.FridgeFreezer.Event.DoorAlarmRefrigerator"
    )
    REFRIGERATION_FRIDGE_FREEZER_EVENT_TEMPERATURE_ALARM_FREEZER = (
        "Refrigeration.FridgeFreezer.Event.TemperatureAlarmFreezer"
    )
    REFRIGERATION_FRIDGE_FREEZER_SETTING_SETPOINT_TEMPERATURE_FREEZER = (
        SettingKey.REFRIGERATION_FRIDGE_FREEZER_SETPOINT_TEMPERATURE_FREEZER
    )
    REFRIGERATION_FRIDGE_FREEZER_SETTING_SETPOINT_TEMPERATURE_REFRIGERATOR = (
        SettingKey.REFRIGERATION_FRIDGE_FREEZER_SETPOINT_TEMPERATURE_REFRIGERATOR
    )
    REFRIGERATION_FRIDGE_FREEZER_SETTING_SUPER_MODE_FREEZER = (
        SettingKey.REFRIGERATION_FRIDGE_FREEZER_SUPER_MODE_FREEZER
    )
    REFRIGERATION_FRIDGE_FREEZER_SETTING_SUPER_MODE_REFRIGERATOR = (
        SettingKey.REFRIGERATION_FRIDGE_FREEZER_SUPER_MODE_REFRIGERATOR
    )
    STOP_PROGRAM = "StopProgram"


class EventType(StrEnum):
    """Represent an event type."""

    KEEP_ALIVE = "KEEP-ALIVE"
    STATUS = "STATUS"
    EVENT = "EVENT"
    NOTIFY = "NOTIFY"
    CONNECTED = "CONNECTED"
    DISCONNECTED = "DISCONNECTED"
    PAIRED = "PAIRED"
    DEPAIRED = "DEPAIRED"



"""Provide common constants for Home Connect API."""

from logging import getLogger

API_ENDPOINT = "https://api.home-connect.com"
OAUTH2_AUTHORIZE = "https://api.home-connect.com/security/oauth/authorize"
OAUTH2_TOKEN = "https://api.home-connect.com/security/oauth/token"  # noqa: S105
LOGGER = getLogger(__package__)



"""Test the client."""

import json

from httpx import AsyncClient, ReadTimeout, RequestError, codes
import pytest
from pytest_httpx import HTTPXMock, IteratorStream

from aiohomeconnect.client import AbstractAuth, Client
from aiohomeconnect.model import ArrayOfEvents, Event, EventKey, EventMessage, EventType
from aiohomeconnect.model.error import (
    EventStreamInterruptedError,
    ForbiddenError,
    HomeConnectApiError,
    HomeConnectRequestError,
    InternalServerError,
    NotAcceptableError,
    RequestTimeoutError,
    TooManyRequestsError,
    UnauthorizedError,
    UnsupportedMediaTypeError,
)

TEST_ACCESS_TOKEN = "1234"
TEST_HA_ID = "SIEMENS-HCS02DWH1-6BE58C26DCC1"
TEST_EVENT_TYPE = EventType.NOTIFY

STREAM_EVENT_CASES = [
    (
        json.dumps(
            {
                "items": [
                    {
                        "handling": "none",
                        "key": EventKey.DISHCARE_DISHWASHER_OPTION_HALF_LOAD.value,
                        "level": "hint",
                        "timestamp": 1733611453,
                        "uri": "/a/random/relative/uri",
                        "value": True,
                    },
                    {
                        "handling": "none",
                        "key": EventKey.BSH_COMMON_OPTION_REMAINING_PROGRAM_TIME.value,
                        "level": "hint",
                        "timestamp": 1733611453,
                        "unit": "seconds",
                        "uri": "/a/random/relative/uri",
                        "value": 13800,
                    },
                ]
            }
        ),
        EventMessage(
            TEST_HA_ID,
            TEST_EVENT_TYPE,
            ArrayOfEvents(
                [
                    Event(
                        handling="none",
                        key=EventKey.DISHCARE_DISHWASHER_OPTION_HALF_LOAD,
                        raw_key=EventKey.DISHCARE_DISHWASHER_OPTION_HALF_LOAD.value,
                        level="hint",
                        timestamp=1733611453,
                        uri="/a/random/relative/uri",
                        value=True,
                    ),
                    Event(
                        handling="none",
                        key=EventKey.BSH_COMMON_OPTION_REMAINING_PROGRAM_TIME,
                        raw_key=EventKey.BSH_COMMON_OPTION_REMAINING_PROGRAM_TIME.value,
                        level="hint",
                        timestamp=1733611453,
                        unit="seconds",
                        uri="/a/random/relative/uri",
                        value=13800,
                    ),
                ]
            ),
        ),
    ),
    (
        json.dumps(
            {
                "items": [
                    {
                        "handling": "acknowledge",
                        "key": EventKey.BSH_COMMON_EVENT_PROGRAM_ABORTED.value,
                        "level": "hint",
                        "timestamp": 1733616930,
                        "value": "BSH.Common.EnumType.EventPresentState.Present",
                    }
                ],
            }
        ),
        EventMessage(
            TEST_HA_ID,
            TEST_EVENT_TYPE,
            ArrayOfEvents(
                [
                    Event(
                        handling="acknowledge",
                        key=EventKey.BSH_COMMON_EVENT_PROGRAM_ABORTED,
                        raw_key=EventKey.BSH_COMMON_EVENT_PROGRAM_ABORTED.value,
                        level="hint",
                        timestamp=1733616930,
                        value="BSH.Common.EnumType.EventPresentState.Present",
                    )
                ]
            ),
        ),
    ),
    (
        json.dumps(
            {
                "handling": "none",
                "key": "BSH.Common.Appliance.Disconnected",
                "level": "hint",
                "timestamp": 1733611817,
                "value": True,
            }
        ),
        EventMessage(
            TEST_HA_ID,
            TEST_EVENT_TYPE,
            ArrayOfEvents(
                [
                    Event(
                        handling="none",
                        key=EventKey.BSH_COMMON_APPLIANCE_DISCONNECTED,
                        raw_key=EventKey.BSH_COMMON_APPLIANCE_DISCONNECTED.value,
                        level="hint",
                        timestamp=1733611817,
                        value=True,
                    )
                ]
            ),
        ),
    ),
    (
        "",
        EventMessage(TEST_HA_ID, TEST_EVENT_TYPE, ArrayOfEvents([])),
    ),
]


class AuthClient(AbstractAuth):
    """Use the abstract auth."""

    async def async_get_access_token(self) -> str:
        """Return a valid access token."""
        return TEST_ACCESS_TOKEN


async def test_abstract_auth_request(
    httpx_client: AsyncClient, httpx_mock: HTTPXMock
) -> None:
    """Test the abstract auth request."""
    url_query = "key1=value1&key2=value2"
    httpx_mock.add_response(
        url=f"https://example.com/api/test?{url_query}",
        json={"test": "test_result"},
    )

    params = {"key1": "value1", "key2": "value2"}
    client = AuthClient(httpx_client, "https://example.com")
    response = await client.request("GET", "/test", params=params)

    assert response.status_code == 200
    assert response.json() == {"test": "test_result"}
    request = httpx_mock.get_request()
    assert request
    assert request.headers["authorization"] == f"Bearer {TEST_ACCESS_TOKEN}"
    assert request.url.query.decode(encoding="utf-8") == "key1=value1&key2=value2"


async def test_abstract_auth_sse(
    httpx_client: AsyncClient, httpx_mock: HTTPXMock
) -> None:
    """Test the abstract auth sse."""
    url_query = "key1=value1&key2=value2"

    httpx_mock.add_response(
        url=f"https://example.com/api/test?{url_query}",
        stream=IteratorStream(
            [
                b'data: {"test": "test_result"}\n\n',
                b'data: {"test": "test_result"}\n\n',
            ]
        ),
        headers={"Content-Type": "text/event-stream"},
    )

    params = {"key1": "value1", "key2": "value2"}
    client = AuthClient(httpx_client, "https://example.com")
    async with client.connect_sse("GET", "/test", params=params) as event_source:
        assert event_source.response.status_code == 200
        num_events = 0
        async for event in event_source.aiter_sse():
            assert event.data == '{"test": "test_result"}'
            num_events += 1
        assert num_events == 2

    request = httpx_mock.get_request()
    assert request
    assert request.headers["authorization"] == f"Bearer {TEST_ACCESS_TOKEN}"
    assert request.url.query.decode(encoding="utf-8") == "key1=value1&key2=value2"


@pytest.mark.parametrize(
    ("status_code", "exception"),
    [
        (codes.UNAUTHORIZED, UnauthorizedError),
        (codes.FORBIDDEN, ForbiddenError),
        (codes.NOT_ACCEPTABLE, NotAcceptableError),
        (codes.REQUEST_TIMEOUT, RequestTimeoutError),
        (codes.UNSUPPORTED_MEDIA_TYPE, UnsupportedMediaTypeError),
        (codes.IM_A_TEAPOT, HomeConnectApiError),
        (codes.INTERNAL_SERVER_ERROR, InternalServerError),
    ],
)
async def test_generic_http_error(
    httpx_client: AsyncClient,
    httpx_mock: HTTPXMock,
    status_code: int,
    exception: type[Exception],
) -> None:
    """Test stream all events http error."""
    httpx_mock.add_response(
        url="https://example.com/api/homeappliances",
        status_code=status_code,
        json={
            "error": {"key": "some_error_key", "description": "some_error_description"}
        },
    )

    client = Client(AuthClient(httpx_client, "https://example.com"))

    with pytest.raises(exception):
        await client.get_home_appliances()


async def test_generic_http_error_no_content(
    httpx_client: AsyncClient,
    httpx_mock: HTTPXMock,
) -> None:
    """Test stream all events http error with no content."""
    httpx_mock.add_response(
        url="https://example.com/api/homeappliances",
        status_code=codes.IM_A_TEAPOT,
    )

    client = Client(AuthClient(httpx_client, "https://example.com"))

    with pytest.raises(HomeConnectApiError):
        await client.get_home_appliances()


async def test_rate_limit_error(
    httpx_client: AsyncClient, httpx_mock: HTTPXMock
) -> None:
    """Test the abstract auth request."""
    httpx_mock.add_response(
        url="https://example.com/api/homeappliances",
        status_code=codes.TOO_MANY_REQUESTS,
        json={
            "error": {"key": "some_error_key", "description": "some_error_description"}
        },
        headers={"Retry-After": "1"},
    )

    client = Client(AuthClient(httpx_client, "https://example.com"))

    with pytest.raises(TooManyRequestsError) as exc_info:
        await client.get_home_appliances()

    assert exc_info.value.key == "some_error_key"
    assert exc_info.value.description == "some_error_description"
    assert exc_info.value.retry_after == 1


async def test_request_error(
    httpx_client: AsyncClient,
    httpx_mock: HTTPXMock,
) -> None:
    """Test stream events from a specific home appliance http error."""
    httpx_mock.add_exception(
        exception=RequestError("some error"),
        url="https://example.com/api/homeappliances",
    )

    client = Client(AuthClient(httpx_client, "https://example.com"))

    with pytest.raises(HomeConnectRequestError):
        await client.get_home_appliances()


@pytest.mark.parametrize(
    ("event_data", "event_message"),
    STREAM_EVENT_CASES,
)
async def test_stream_all_events(
    httpx_client: AsyncClient,
    httpx_mock: HTTPXMock,
    event_data: str,
    event_message: EventMessage,
) -> None:
    """Test stream all events."""
    httpx_mock.add_response(
        url="https://example.com/api/homeappliances/events",
        stream=IteratorStream(
            [
                "\n".join(
                    [
                        f"id: {TEST_HA_ID}",
                        f"data: {event_data}",
                        f"event: {TEST_EVENT_TYPE}",
                        "\n",
                    ]
                ).encode()
            ]
        ),
        headers={"Content-Type": "text/event-stream"},
    )

    client = Client(AuthClient(httpx_client, "https://example.com"))

    assert await anext(client.stream_all_events()) == event_message


@pytest.mark.parametrize(
    ("status_code", "exception"),
    [
        (codes.UNAUTHORIZED, UnauthorizedError),
        (codes.FORBIDDEN, ForbiddenError),
        (codes.NOT_ACCEPTABLE, NotAcceptableError),
        (codes.IM_A_TEAPOT, HomeConnectApiError),
        (codes.TOO_MANY_REQUESTS, TooManyRequestsError),
        (codes.INTERNAL_SERVER_ERROR, InternalServerError),
    ],
)
async def test_stream_all_events_http_error(
    httpx_client: AsyncClient,
    httpx_mock: HTTPXMock,
    status_code: int,
    exception: type[Exception],
) -> None:
    """Test stream all events http error."""
    httpx_mock.add_response(
        url="https://example.com/api/homeappliances/events",
        status_code=status_code,
        json={
            "error": {"key": "some_error_key", "description": "some_error_description"}
        },
    )

    client = Client(AuthClient(httpx_client, "https://example.com"))

    with pytest.raises(exception):
        await anext(client.stream_all_events())


@pytest.mark.parametrize(
    ("event_data", "event_message"),
    STREAM_EVENT_CASES,
)
async def test_stream_events(
    httpx_client: AsyncClient,
    httpx_mock: HTTPXMock,
    event_data: str,
    event_message: EventMessage,
) -> None:
    """Test stream events from a specific home appliance."""
    httpx_mock.add_response(
        url=f"https://example.com/api/homeappliances/{TEST_HA_ID}/events",
        stream=IteratorStream(
            [
                "\n".join(
                    [
                        f"id: {TEST_HA_ID}",
                        f"data: {event_data}",
                        f"event: {TEST_EVENT_TYPE}",
                        "\n",
                    ]
                ).encode()
            ]
        ),
        headers={"Content-Type": "text/event-stream"},
    )

    client = Client(AuthClient(httpx_client, "https://example.com"))

    assert await anext(client.stream_events(TEST_HA_ID)) == event_message


@pytest.mark.parametrize(
    ("status_code", "exception"),
    [
        (401, UnauthorizedError),
        (403, ForbiddenError),
        (406, NotAcceptableError),
        (418, HomeConnectApiError),
        (429, TooManyRequestsError),
        (500, InternalServerError),
    ],
)
async def test_stream_events_http_error(
    httpx_client: AsyncClient,
    httpx_mock: HTTPXMock,
    status_code: int,
    exception: type[Exception],
) -> None:
    """Test stream events from a specific home appliance http error."""
    httpx_mock.add_response(
        url=f"https://example.com/api/homeappliances/{TEST_HA_ID}/events",
        status_code=status_code,
        json={
            "error": {"key": "some_error_key", "description": "some_error_description"}
        },
    )

    client = Client(AuthClient(httpx_client, "https://example.com"))

    with pytest.raises(exception):
        await anext(client.stream_events(TEST_HA_ID))


@pytest.mark.parametrize(
    ("exception", "expected_exception"),
    [
        (ReadTimeout, EventStreamInterruptedError),
        (RequestError, HomeConnectRequestError),
    ],
)
async def test_stream_events_request_error(
    httpx_client: AsyncClient,
    httpx_mock: HTTPXMock,
    exception: type[Exception],
    expected_exception: type[Exception],
) -> None:
    """Test stream events from a specific home appliance http error."""
    httpx_mock.add_exception(
        exception=exception("some error"),
        url="https://example.com/api/homeappliances/events",
    )

    client = Client(AuthClient(httpx_client, "https://example.com"))

    with pytest.raises(expected_exception):
        await anext(client.stream_all_events())



"""Provide a client for Home Connect API."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator, AsyncIterator
from contextlib import asynccontextmanager
from typing import Any

from httpx import (
    AsyncClient,
    ReadTimeout,
    RemoteProtocolError,
    RequestError,
    Response,
    Timeout,
    codes,
)
from httpx_sse import EventSource, aconnect_sse

from aiohomeconnect.model import EventMessage, EventType

from .const import LOGGER
from .model import (
    ArrayOfAvailablePrograms,
    ArrayOfCommands,
    ArrayOfHomeAppliances,
    ArrayOfImages,
    ArrayOfOptions,
    ArrayOfPrograms,
    ArrayOfSettings,
    ArrayOfStatus,
    CommandKey,
    GetSetting,
    HomeAppliance,
    Language,
    Option,
    OptionKey,
    Program,
    ProgramDefinition,
    ProgramKey,
    PutCommand,
    PutCommands,
    PutSetting,
    PutSettings,
    SettingKey,
    Status,
    StatusKey,
)
from .model.error import (
    ActiveProgramNotSetError,
    Conflict,
    ConflictError,
    EventStreamInterruptedError,
    ForbiddenError,
    HomeConnectApiError,
    HomeConnectRequestError,
    InternalServerError,
    NoProgramActiveError,
    NoProgramSelectedError,
    NotAcceptableError,
    NotFoundError,
    ProgramNotAvailableError,
    RequestTimeoutError,
    SelectedProgramNotSetError,
    TooManyRequestsError,
    UnauthorizedError,
    UnsupportedMediaTypeError,
    WrongOperationStateError,
)


def _raise_generic_error(response: Response) -> None:
    """Raise a generic error if the response is an error."""
    if response.is_error:
        raise (
            HomeConnectApiError.from_dict(error)
            if response.content and (error := response.json().get("error"))
            else HomeConnectApiError(
                "unknown",
                f"Unknown HTTP error (Status code: {response.status_code})",
            )
        )


class AbstractAuth(ABC):
    """Abstract class to make authenticated requests."""

    def __init__(self, httpx_client: AsyncClient, host: str) -> None:
        """Initialize the auth."""
        self.client = httpx_client
        self.host = host

    @abstractmethod
    async def async_get_access_token(self) -> str:
        """Return a valid access token."""

    async def _get_headers(self, headers: dict[str, str] | None) -> dict[str, str]:
        """Return the headers for the request."""
        headers = {} if headers is None else dict(headers)
        headers = {key: val for key, val in headers.items() if val is not None}

        access_token = await self.async_get_access_token()
        headers["authorization"] = f"Bearer {access_token}"
        return headers

    async def request(self, method: str, url: str, **kwargs: Any) -> Response:
        """Make a request.

        The url parameter must start with a slash.
        """
        headers = await self._get_headers(kwargs.pop("headers", None))
        headers["accept"] = "application/vnd.bsh.sdk.v1+json"
        data = kwargs.pop("data", None)
        if data is not None:
            headers["content-type"] = "application/vnd.bsh.sdk.v1+json"

        LOGGER.debug("Request: %s %s | Data: %s", method, url, data)

        try:
            response = await self.client.request(
                method,
                f"{self.host}/api{url}",
                **kwargs,
                headers=headers,
                json={"data": data} if data is not None else None,
                timeout=Timeout(5, read=30),
            )
        except RequestError as e:
            raise HomeConnectRequestError(f"{type(e).__name__}: {e}") from e

        LOGGER.debug("Response: \n%s", response.text)

        match response.status_code:
            case codes.UNAUTHORIZED:
                raise UnauthorizedError.from_dict(response.json()["error"])
            case codes.FORBIDDEN:
                raise ForbiddenError.from_dict(response.json()["error"])
            case codes.NOT_ACCEPTABLE:
                raise NotAcceptableError.from_dict(response.json()["error"])
            case codes.REQUEST_TIMEOUT:
                raise RequestTimeoutError.from_dict(response.json()["error"])
            case codes.UNSUPPORTED_MEDIA_TYPE:
                raise UnsupportedMediaTypeError.from_dict(response.json()["error"])
            case codes.TOO_MANY_REQUESTS:
                err = TooManyRequestsError.from_dict(response.json()["error"])
                retry_after = response.headers.get("Retry-After")
                err.retry_after = int(retry_after) if retry_after else None
                raise err
            case codes.INTERNAL_SERVER_ERROR:
                raise InternalServerError.from_dict(response.json()["error"])
            case _:
                return response

    @asynccontextmanager
    async def connect_sse(
        self,
        method: str,
        url: str,
        **kwargs: Any,
    ) -> AsyncIterator[EventSource]:
        """Create a SSE connection."""
        headers = await self._get_headers(kwargs.pop("headers", None))

        try:
            async with aconnect_sse(
                self.client,
                method,
                f"{self.host}/api{url}",
                **kwargs,
                headers=headers,
            ) as event_source:
                yield event_source
        except (ReadTimeout, RemoteProtocolError) as e:
            raise EventStreamInterruptedError(f"{type(e).__name__}: {e}") from e
        except RequestError as e:
            raise HomeConnectRequestError(f"{type(e).__name__}: {e}") from e


class Client:
    """Represent a client for the Home Connect API."""

    def __init__(self, auth: AbstractAuth) -> None:
        """Initialize the client."""
        self._auth = auth

    async def get_home_appliances(self) -> ArrayOfHomeAppliances:
        """Get all home appliances which are paired with the logged-in user account.

        This endpoint returns a list of all home appliances which are paired
        with the logged-in user account. All paired home appliances are returned
        independent of their current connection state. The connection state can
        be retrieved within the field 'connected' of the respective home appliance.
        The haId is the primary access key for further API access to a specific
        home appliance.
        """
        response = await self._auth.request(
            "GET",
            "/homeappliances",
            headers=None,
        )
        _raise_generic_error(response)
        return ArrayOfHomeAppliances.from_dict(response.json()["data"])

    async def get_specific_appliance(
        self,
        ha_id: str,
    ) -> HomeAppliance:
        """Get a specific paired home appliance.

        This endpoint returns a specific home appliance which is paired with the
        logged-in user account. It is returned independent of their current
        connection state. The connection state can be retrieved within the field
        'connected' of the respective home appliance.
        The haId is the primary access key for further API access to a specific
        home appliance.
        """
        response = await self._auth.request(
            "GET",
            f"/homeappliances/{ha_id}",
            headers=None,
        )
        _raise_generic_error(response)
        return HomeAppliance.from_dict(response.json()["data"])

    async def get_all_programs(
        self,
        ha_id: str,
        *,
        accept_language: Language | None = Language.EN,
    ) -> ArrayOfPrograms:
        """Get all programs of a given home appliance."""
        response = await self._auth.request(
            "GET",
            f"/homeappliances/{ha_id}/programs",
            headers={"Accept-Language": accept_language},
        )
        match response.status_code:
            case codes.CONFLICT:
                raise Conflict.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)
        return ArrayOfPrograms.from_dict(response.json()["data"])

    async def get_available_programs(
        self,
        ha_id: str,
        *,
        accept_language: Language | None = Language.EN,
    ) -> ArrayOfAvailablePrograms:
        """Get all currently available programs on the given home appliance."""
        response = await self._auth.request(
            "GET",
            f"/homeappliances/{ha_id}/programs/available",
            headers={"Accept-Language": accept_language},
        )
        match response.status_code:
            case codes.CONFLICT:
                raise WrongOperationStateError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)
        return ArrayOfAvailablePrograms.from_dict(response.json()["data"])

    async def get_available_program(
        self,
        ha_id: str,
        *,
        program_key: ProgramKey,
        accept_language: Language | None = Language.EN,
    ) -> ProgramDefinition:
        """Get a specific available program."""
        response = await self._auth.request(
            "GET",
            f"/homeappliances/{ha_id}/programs/available/{program_key}",
            headers={"Accept-Language": accept_language},
        )
        match response.status_code:
            case codes.CONFLICT:
                raise ProgramNotAvailableError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)
        return ProgramDefinition.from_dict(response.json()["data"])

    async def get_active_program(
        self,
        ha_id: str,
        *,
        accept_language: Language | None = Language.EN,
    ) -> Program:
        """Get the active program."""
        response = await self._auth.request(
            "GET",
            f"/homeappliances/{ha_id}/programs/active",
            headers={"Accept-Language": accept_language},
        )
        match response.status_code:
            case codes.NOT_FOUND:
                raise NoProgramActiveError.from_dict(response.json()["error"])
            case codes.CONFLICT:
                raise ConflictError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)
        return Program.from_dict(response.json()["data"])

    async def start_program(
        self,
        ha_id: str,
        *,
        program_key: ProgramKey,
        name: str | None = None,
        options: list[Option] | None = None,
        accept_language: Language | None = Language.EN,
    ) -> None:
        """Start the given program.

        By putting a program object to this endpoint, the system will try to
        start it if all preconditions are fulfilled:
        * Home appliance is connected
        * *Remote Control* and *Remote Control Start Allowed* is enabled
        * No other program is currently active

        Furthermore, the program must exist on the home appliance and its
        options must be set correctly.
        Keys of program, which can be executed on an oven, are for instance:
        * *Cooking.Oven.Program.HeatingMode.HotAir*
        * *Cooking.Oven.Program.HeatingMode.TopBottomHeating*
        * *Cooking.Oven.Program.HeatingMode.PizzaSetting*
        * *Cooking.Oven.Program.HeatingMode.PreHeating*

        Keys for options of these oven programs are:
        * *Cooking.Oven.Option.SetpointTemperature*: 30 - 250 °C
        * *BSH.Common.Option.Duration*: 1 - 86340 seconds

        For further documentation, visit the appliance-specific programs pages:
        * [Cleaning Robot](https://api-docs.home-connect.com/programs-and-options?#cleaning-robot)
        * [Coffee Machine](https://api-docs.home-connect.com/programs-and-options?#coffee-machine)
        * [Cooktop](https://api-docs.home-connect.com/programs-and-options?#cooktop)
        * [Cook Processor](https://api-docs.home-connect.com/programs-and-options?#cook-processor)
        * [Dishwasher](https://api-docs.home-connect.com/programs-and-options?#dishwasher)
        * [Dryer](https://api-docs.home-connect.com/programs-and-options?#dryer)
        * [Hood](https://api-docs.home-connect.com/programs-and-options?#hood)
        * [Oven](https://api-docs.home-connect.com/programs-and-options?#oven)
        * [Warming Drawer](https://api-docs.home-connect.com/programs-and-options?#warming-drawer)
        * [Washer](https://api-docs.home-connect.com/programs-and-options?#washer)
        * [Washer Dryer](https://api-docs.home-connect.com/programs-and-options?#washer-dryer)

        There are no programs available for freezers, fridge freezers,
        refrigerators and wine coolers.
        """
        program = Program(key=program_key, name=name, options=options)
        response = await self._auth.request(
            "PUT",
            f"/homeappliances/{ha_id}/programs/active",
            headers={"Accept-Language": accept_language},
            data=program.to_dict(),
        )
        match response.status_code:
            case codes.CONFLICT:
                raise ConflictError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)

    async def stop_program(
        self,
        ha_id: str,
        *,
        accept_language: Language | None = Language.EN,
    ) -> None:
        """Stop the active program."""
        response = await self._auth.request(
            "DELETE",
            f"/homeappliances/{ha_id}/programs/active",
            headers={"Accept-Language": accept_language},
        )
        match response.status_code:
            case codes.CONFLICT:
                raise WrongOperationStateError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)

    async def get_active_program_options(
        self,
        ha_id: str,
        *,
        accept_language: Language | None = Language.EN,
    ) -> ArrayOfOptions:
        """Get all options of the active program.

        You can retrieve a list of options of the currently running program.

        For detailed documentation of the available options,
        visit the appliance-specific programs pages:
        * [Cleaning Robot](https://api-docs.home-connect.com/programs-and-options?#cleaning-robot)
        * [Coffee Machine](https://api-docs.home-connect.com/programs-and-options?#coffee-machine)
        * [Cooktop](https://api-docs.home-connect.com/programs-and-options?#cooktop)
        * [Cook Processor](https://api-docs.home-connect.com/programs-and-options?#cook-processor)
        * [Dishwasher](https://api-docs.home-connect.com/programs-and-options?#dishwasher)
        * [Dryer](https://api-docs.home-connect.com/programs-and-options?#dryer)
        * [Hood](https://api-docs.home-connect.com/programs-and-options?#hood)
        * [Oven](https://api-docs.home-connect.com/programs-and-options?#oven)
        * [Warming Drawer](https://api-docs.home-connect.com/programs-and-options?#warming-drawer)
        * [Washer](https://api-docs.home-connect.com/programs-and-options?#washer)
        * [Washer Dryer](https://api-docs.home-connect.com/programs-and-options?#washer-dryer)
        """
        response = await self._auth.request(
            "GET",
            f"/homeappliances/{ha_id}/programs/active/options",
            headers={"Accept-Language": accept_language},
        )
        match response.status_code:
            case codes.NOT_FOUND:
                raise NoProgramActiveError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)
        return ArrayOfOptions.from_dict(response.json()["data"])

    async def set_active_program_options(
        self,
        ha_id: str,
        *,
        array_of_options: ArrayOfOptions,
        accept_language: Language | None = Language.EN,
    ) -> None:
        """Set all options of the active program.

        Update the options for the currently running program.
        With this API endpoint, you have to provide all options with
        their new values. If you want to update only one option, you can use the
        endpoint specific to that option.

        Please note that changing options of the running program is currently only
        supported by ovens.
        """
        response = await self._auth.request(
            "PUT",
            f"/homeappliances/{ha_id}/programs/active/options",
            headers={"Accept-Language": accept_language},
            data=array_of_options.to_dict(),
        )
        match response.status_code:
            case codes.CONFLICT:
                raise ActiveProgramNotSetError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)

    async def get_active_program_option(
        self,
        ha_id: str,
        *,
        option_key: OptionKey,
        accept_language: Language | None = Language.EN,
    ) -> Option:
        """Get a specific option of the active program."""
        response = await self._auth.request(
            "GET",
            f"/homeappliances/{ha_id}/programs/active/options/{option_key}",
            headers={"Accept-Language": accept_language},
        )
        match response.status_code:
            case codes.NOT_FOUND:
                raise NoProgramActiveError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)
        return Option.from_dict(response.json()["data"])

    async def set_active_program_option(
        self,
        ha_id: str,
        *,
        option_key: OptionKey,
        value: Any,
        name: str | None = None,
        display_value: str | None = None,
        unit: str | None = None,
        accept_language: Language | None = Language.EN,
    ) -> None:
        """Set a specific option of the active program.

        This operation can be used to modify one specific option of the active
        program, e.g. to extend the duration of the active program by
        another 5 minutes.

        Please note that changing options of the running program is currently only
        supported by ovens.
        """
        option = Option(
            key=option_key,
            name=name,
            value=value,
            display_value=display_value,
            unit=unit,
        )
        response = await self._auth.request(
            "PUT",
            f"/homeappliances/{ha_id}/programs/active/options/{option_key}",
            headers={"Accept-Language": accept_language},
            data=option.to_dict(),
        )
        match response.status_code:
            case codes.CONFLICT:
                raise ActiveProgramNotSetError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)

    async def get_selected_program(
        self,
        ha_id: str,
        *,
        accept_language: Language | None = Language.EN,
    ) -> Program:
        """Get the selected program.

        In most cases the selected program is the program which is currently
        shown on the display of the home appliance. This program can then be
        manually adjusted or started on the home appliance itself.
        """
        response = await self._auth.request(
            "GET",
            f"/homeappliances/{ha_id}/programs/selected",
            headers={"Accept-Language": accept_language},
        )
        match response.status_code:
            case codes.NOT_FOUND:
                raise NoProgramSelectedError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)
        return Program.from_dict(response.json()["data"])

    async def set_selected_program(
        self,
        ha_id: str,
        *,
        program_key: ProgramKey,
        name: str | None = None,
        options: list[Option] | None = None,
        accept_language: Language | None = Language.EN,
    ) -> None:
        """Select the given program.

        In most cases the selected program is the program which is currently
        shown on the display of the home appliance. This program can then be
        manually adjusted or started on the home appliance itself.

        A selected program will not be started automatically. You don't have
        to set a program as selected first if you intend to start it via API -
        you can set it directly as active program.

        Selecting a program will update the available options and constraints
        directly from the home appliance. Any changes to the available options
        due to the state of the appliance is only reflected in the selected program.
        """
        program = Program(key=program_key, name=name, options=options)
        response = await self._auth.request(
            "PUT",
            f"/homeappliances/{ha_id}/programs/selected",
            headers={"Accept-Language": accept_language},
            data=program.to_dict(),
        )
        match response.status_code:
            case codes.CONFLICT:
                raise ConflictError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)

    async def get_selected_program_options(
        self,
        ha_id: str,
        *,
        accept_language: Language | None = Language.EN,
    ) -> ArrayOfOptions:
        """Get all options of the selected program."""
        response = await self._auth.request(
            "GET",
            f"/homeappliances/{ha_id}/programs/selected/options",
            headers={"Accept-Language": accept_language},
        )
        match response.status_code:
            case codes.NOT_FOUND:
                raise NoProgramSelectedError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)
        return ArrayOfOptions.from_dict(response.json()["data"])

    async def set_selected_program_options(
        self,
        ha_id: str,
        *,
        array_of_options: ArrayOfOptions,
        accept_language: Language | None = Language.EN,
    ) -> None:
        """Set all options of the selected program."""
        response = await self._auth.request(
            "PUT",
            f"/homeappliances/{ha_id}/programs/selected/options",
            headers={"Accept-Language": accept_language},
            data=array_of_options.to_dict(),
        )
        match response.status_code:
            case codes.CONFLICT:
                raise SelectedProgramNotSetError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)

    async def get_selected_program_option(
        self,
        ha_id: str,
        *,
        option_key: OptionKey,
        accept_language: Language | None = Language.EN,
    ) -> Option:
        """Get a specific option of the selected program."""
        response = await self._auth.request(
            "GET",
            f"/homeappliances/{ha_id}/programs/selected/options/{option_key}",
            headers={"Accept-Language": accept_language},
        )
        match response.status_code:
            case codes.NOT_FOUND:
                raise NoProgramSelectedError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)
        return Option.from_dict(response.json()["data"])

    async def set_selected_program_option(
        self,
        ha_id: str,
        *,
        option_key: OptionKey,
        value: Any,
        name: str | None = None,
        display_value: str | None = None,
        unit: str | None = None,
        accept_language: Language | None = Language.EN,
    ) -> None:
        """Set a specific option of the selected program."""
        option = Option(
            key=option_key,
            name=name,
            value=value,
            display_value=display_value,
            unit=unit,
        )
        response = await self._auth.request(
            "PUT",
            f"/homeappliances/{ha_id}/programs/selected/options/{option_key}",
            headers={"Accept-Language": accept_language},
            data=option.to_dict(),
        )
        match response.status_code:
            case codes.CONFLICT:
                raise SelectedProgramNotSetError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)

    async def get_images(
        self,
        ha_id: str,
        *,
        accept_language: Language | None = Language.EN,
    ) -> ArrayOfImages:
        """Get a list of available images."""
        response = await self._auth.request(
            "GET",
            f"/homeappliances/{ha_id}/images",
            headers={"Accept-Language": accept_language},
        )
        _raise_generic_error(response)
        return ArrayOfImages.from_dict(response.json()["data"])

    async def get_image(
        self,
        ha_id: str,
        *,
        image_key: str,
    ) -> None:
        """Get a specific image."""
        response = await self._auth.request(
            "GET",
            f"/homeappliances/{ha_id}/images/{image_key}",
            headers=None,
        )
        match response.status_code:
            case codes.NOT_FOUND:
                raise NotFoundError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)

    async def get_settings(
        self,
        ha_id: str,
        *,
        accept_language: Language | None = Language.EN,
    ) -> ArrayOfSettings:
        """Get a list of available settings.

        Get a list of available setting of the home appliance.

        Further documentation
        can be found [here](https://api-docs.home-connect.com/settings).
        """
        response = await self._auth.request(
            "GET",
            f"/homeappliances/{ha_id}/settings",
            headers={"Accept-Language": accept_language},
        )
        _raise_generic_error(response)
        return ArrayOfSettings.from_dict(response.json()["data"])

    async def set_settings(
        self,
        ha_id: str,
        *,
        put_settings: PutSettings,
        accept_language: Language | None = Language.EN,
    ) -> None:
        """Set multiple settings."""
        response = await self._auth.request(
            "PUT",
            f"/homeappliances/{ha_id}/settings",
            headers={"Accept-Language": accept_language},
            data=put_settings.to_dict(),
        )
        match response.status_code:
            case codes.CONFLICT:
                raise ConflictError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)

    async def get_setting(
        self,
        ha_id: str,
        *,
        setting_key: SettingKey,
        accept_language: Language | None = Language.EN,
    ) -> GetSetting:
        """Get a specific setting."""
        response = await self._auth.request(
            "GET",
            f"/homeappliances/{ha_id}/settings/{setting_key}",
            headers={"Accept-Language": accept_language},
        )
        match response.status_code:
            case codes.NOT_FOUND:
                raise NotFoundError.from_dict(response.json()["error"])
            case codes.CONFLICT:
                raise ConflictError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)
        return GetSetting.from_dict(response.json()["data"])

    async def set_setting(
        self,
        ha_id: str,
        *,
        setting_key: SettingKey,
        value: Any,
        accept_language: Language | None = Language.EN,
    ) -> None:
        """Set a specific setting."""
        put_setting = PutSetting(key=setting_key, value=value)
        response = await self._auth.request(
            "PUT",
            f"/homeappliances/{ha_id}/settings/{setting_key}",
            headers={"Accept-Language": accept_language},
            data=put_setting.to_dict(),
        )
        match response.status_code:
            case codes.NOT_FOUND:
                raise NotFoundError.from_dict(response.json()["error"])
            case codes.CONFLICT:
                raise ConflictError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)

    async def get_status(
        self,
        ha_id: str,
        *,
        accept_language: Language | None = Language.EN,
    ) -> ArrayOfStatus:
        """Get a list of available status.

        A detailed description of the available status
        can be found [here](https://api-docs.home-connect.com/states).
        """
        response = await self._auth.request(
            "GET",
            f"/homeappliances/{ha_id}/status",
            headers={"Accept-Language": accept_language},
        )
        match response.status_code:
            case codes.CONFLICT:
                raise ConflictError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)
        return ArrayOfStatus.from_dict(response.json()["data"])

    async def get_status_value(
        self,
        ha_id: str,
        *,
        status_key: StatusKey,
        accept_language: Language | None = Language.EN,
    ) -> Status:
        """Get a specific status.

        A detailed description of the available status
        can be found [here](https://api-docs.home-connect.com/states).
        """
        response = await self._auth.request(
            "GET",
            f"/homeappliances/{ha_id}/status/{status_key}",
            headers={"Accept-Language": accept_language},
        )
        match response.status_code:
            case codes.NOT_FOUND:
                raise NotFoundError.from_dict(response.json()["error"])
            case codes.CONFLICT:
                raise ConflictError.from_dict(response.json()["error"])
            case _:
                _raise_generic_error(response)
        return Status.from_dict(response.json()["data"])

    async def get_available_commands(
        self,
        ha_id: str,
        *,
        accept_language: Language | None = Language.EN,
    ) -> ArrayOfCommands:
        """Get a list of available and writable commands."""
        response = await self._auth.request(
            "GET",
            f"/homeappliances/{ha_id}/commands",
            headers={"Accept-Language": accept_language},
        )
        _raise_generic_error(response)
        return ArrayOfCommands.from_dict(response.json()["data"])

    async def put_commands(
        self,
        ha_id: str,
        *,
        put_commands: PutCommands,
        accept_language: Language | None = Language.EN,
    ) -> None:
        """Execute multiple commands."""
        response = await self._auth.request(
            "PUT",
            f"/homeappliances/{ha_id}/commands",
            headers={"Accept-Language": accept_language},
            data=put_commands.to_dict(),
        )
        _raise_generic_error(response)

    async def put_command(
        self,
        ha_id: str,
        *,
        command_key: CommandKey,
        value: Any,
        accept_language: Language | None = Language.EN,
    ) -> None:
        """Execute a specific command."""
        put_command = PutCommand(key=command_key, value=value)
        response = await self._auth.request(
            "PUT",
            f"/homeappliances/{ha_id}/commands/{command_key}",
            headers={"Accept-Language": accept_language},
            data=put_command.to_dict(),
        )
        _raise_generic_error(response)

    async def stream_all_events(
        self,
        *,
        accept_language: Language | None = Language.EN,
    ) -> AsyncGenerator[EventMessage]:
        """Get stream of events for all appliances.

        Server Sent Events are available as Eventsource API in JavaScript
        and are implemented by various HTTP client libraries and tools
        including curl.

        Unfortunately, SSE is not compatible to OpenAPI specs and can therefore
        not be properly specified within this API description.

        An SSE event contains three parts separated by linebreaks: event, data and id.
        Different events are separated by empty lines.

        The event field can be one of these types:
        KEEP-ALIVE, STATUS, EVENT, NOTIFY, CONNECTED, DISCONNECTED, PAIRED, DEPAIRED.

        In case of all event types (except KEEP-ALIVE),
        the "data" field is populated with the JSON object defined below.

        The id contains the home appliance ID. (except for KEEP-ALIVE event type)

        Further documentation can be found here:
        * [Events availability matrix](https://api-docs.home-connect.com/events#availability-matrix)
        * [Program changes](https://api-docs.home-connect.com/events#program-changes)
        * [Option changes](https://api-docs.home-connect.com/events#option-changes)
        * [Program progress changes](https://api-docs.home-connect.com/events#program-progress-changes)
        * [Home appliance state changes](https://api-docs.home-connect.com/events#home-appliance-state-changes)
        """
        # We use 60 seconds timeout because at least every 55 seconds a KEEP-ALIVE event
        # will be sent. See https://api-docs.home-connect.com/events/#availability-matrix
        async with self._auth.connect_sse(
            "GET",
            "/homeappliances/events",
            timeout=Timeout(60),
            headers={
                "Accept-Language": accept_language,
            },
        ) as event_source:
            response = event_source.response
            if response.is_error:
                await response.aread()
                match event_source.response.status_code:
                    case codes.UNAUTHORIZED:
                        raise UnauthorizedError.from_dict(response.json()["error"])
                    case codes.FORBIDDEN:
                        raise ForbiddenError.from_dict(response.json()["error"])
                    case codes.NOT_ACCEPTABLE:
                        raise NotAcceptableError.from_dict(response.json()["error"])
                    case codes.TOO_MANY_REQUESTS:
                        raise TooManyRequestsError.from_dict(response.json()["error"])
                    case codes.INTERNAL_SERVER_ERROR:
                        raise InternalServerError.from_dict(response.json()["error"])
                    case _:
                        _raise_generic_error(response)

            async for sse in event_source.aiter_sse():
                LOGGER.debug("Event: %s", sse)

                if (
                    # _value2member_map_ is required for Python 3.11,
                    # remove after dropping support for it.
                    sse.event not in EventType._value2member_map_
                    or sse.event == EventType.KEEP_ALIVE
                ):
                    continue
                yield EventMessage.from_server_sent_event(sse)

    async def stream_events(
        self,
        ha_id: str,
        *,
        accept_language: Language | None = Language.EN,
    ) -> AsyncGenerator[EventMessage]:
        """Get stream of events for one appliance.

        If you want to do a one-time query of the current status, you can ask for
        the content-type `application/vnd.bsh.sdk.v1+json` and get the status
        as normal HTTP response.

        If you want an ongoing stream of events in real time, ask for the content
        type `text/event-stream` and you'll get a stream as Server Sent Events.

        Server Sent Events are available as Eventsource API in JavaScript
        and are implemented by various HTTP client libraries and tools
        including curl.

        Unfortunately, SSE is not compatible to OpenAPI specs and can therefore
        not be properly specified within this API description.

        An SSE event contains three parts separated by linebreaks: event, data and id.
        Different events are separated by empty lines.

        The event field can be one of these types:
        KEEP-ALIVE, STATUS, EVENT, NOTIFY, CONNECTED, DISCONNECTED.

        In case of all event types (except KEEP-ALIVE),
        the "data" field is populated with the JSON object defined below.

        The id contains the home appliance ID.

        Further documentation can be found here:
        * [Events availability matrix](https://api-docs.home-connect.com/events#availability-matrix)
        * [Program changes](https://api-docs.home-connect.com/events#program-changes)
        * [Option changes](https://api-docs.home-connect.com/events#option-changes)
        * [Program progress changes](https://api-docs.home-connect.com/events#program-progress-changes)
        * [Home appliance state changes](https://api-docs.home-connect.com/events#home-appliance-state-changes)
        """
        # We use 60 seconds timeout because at least every 55 seconds a KEEP-ALIVE event
        # will be sent. See https://api-docs.home-connect.com/events/#availability-matrix
        async with self._auth.connect_sse(
            "GET",
            f"/homeappliances/{ha_id}/events",
            timeout=Timeout(60),
            headers={
                "Accept-Language": accept_language,
            },
        ) as event_source:
            response = event_source.response
            if response.is_error:
                await response.aread()
                match event_source.response.status_code:
                    case codes.UNAUTHORIZED:
                        raise UnauthorizedError.from_dict(response.json()["error"])
                    case codes.FORBIDDEN:
                        raise ForbiddenError.from_dict(response.json()["error"])
                    case codes.NOT_ACCEPTABLE:
                        raise NotAcceptableError.from_dict(response.json()["error"])
                    case codes.TOO_MANY_REQUESTS:
                        raise TooManyRequestsError.from_dict(response.json()["error"])
                    case codes.INTERNAL_SERVER_ERROR:
                        raise InternalServerError.from_dict(response.json()["error"])
                    case _:
                        _raise_generic_error(response)

            async for sse in event_source.aiter_sse():
                if (
                    # _value2member_map_ is required for Python 3.11,
                    # remove after dropping support for it.
                    sse.event not in EventType._value2member_map_
                    or sse.event == EventType.KEEP_ALIVE
                ):
                    continue
                yield EventMessage.from_server_sent_event(sse)



"""Provide a model for the Home Connect API."""

from enum import StrEnum

from .appliance import (
    ArrayOfHomeAppliances,
    HomeAppliance,
)
from .command import (
    ArrayOfCommands,
    CommandKey,
    PutCommand,
    PutCommands,
)
from .event import (
    ArrayOfEvents,
    Event,
    EventKey,
    EventMessage,
    EventType,
)
from .image import (
    ArrayOfImages,
)
from .program import (
    ArrayOfAvailablePrograms,
    ArrayOfOptions,
    ArrayOfPrograms,
    Option,
    OptionKey,
    Program,
    ProgramConstraints,
    ProgramDefinition,
    ProgramKey,
)
from .setting import (
    ArrayOfSettings,
    GetSetting,
    PutSetting,
    PutSettings,
    SettingKey,
)
from .status import (
    ArrayOfStatus,
    Status,
    StatusKey,
)

__all__ = [
    "ArrayOfAvailablePrograms",
    "ArrayOfCommands",
    "ArrayOfEvents",
    "ArrayOfHomeAppliances",
    "ArrayOfImages",
    "ArrayOfOptions",
    "ArrayOfPrograms",
    "ArrayOfSettings",
    "ArrayOfStatus",
    "CommandKey",
    "Event",
    "EventKey",
    "EventMessage",
    "EventType",
    "GetSetting",
    "HomeAppliance",
    "Option",
    "OptionKey",
    "Program",
    "ProgramConstraints",
    "ProgramDefinition",
    "ProgramKey",
    "PutCommand",
    "PutCommands",
    "PutSetting",
    "PutSettings",
    "SettingKey",
    "Status",
    "StatusKey",
]


class Language(StrEnum):
    """Represent the language for the response."""

    DE = "de-DE"
    EN = "en-US"
    EN_GB = "en-GB"
